use std::{borrow::Cow, convert::Infallible};

use axum::{http::StatusCode, response::IntoResponse};

type BoxError = Box<dyn std::error::Error>;

pub struct GeneralError {
    pub status_code: StatusCode,
    pub message: Cow<'static, str>,
    pub error: Option<BoxError>,
}

pub type GeneralResult<T> = Result<T, GeneralError>;

impl GeneralError {
    pub fn message(status_code: StatusCode, message: impl Into<Cow<'static, str>>) -> Self {
        Self {
            status_code,
            message: message.into(),
            error: None,
        }
    }

    #[allow(clippy::needless_pass_by_value)]
    pub fn validation(error: validator::ValidationErrors) -> Self {
        Self::message(StatusCode::BAD_REQUEST, error.to_string())
    }

    pub const fn item_not_found() -> Self {
        GeneralError {
            status_code: StatusCode::NOT_FOUND,
            message: Cow::Borrowed("Item not found"),
            error: None,
        }
    }
}

impl IntoResponse for GeneralError {
    fn into_response(self) -> axum::response::Response {
        if let Some(err) = self.error {
            tracing::error!(err, message = %self.message, status_code = ?self.status_code);
        }
        (self.status_code, self.message).into_response()
    }
}

pub trait ErrorHandlingExt<T, E>
where
    Self: Sized,
{
    fn handle(
        self,
        status_code: StatusCode,
        message: impl Into<Cow<'static, str>>,
    ) -> GeneralResult<T>;

    fn handle_internal(self, message: impl Into<Cow<'static, str>>) -> GeneralResult<T> {
        self.handle(StatusCode::INTERNAL_SERVER_ERROR, message)
    }
}

impl<T, E: Into<BoxError>> ErrorHandlingExt<T, E> for Result<T, E> {
    fn handle(
        self,
        status_code: StatusCode,
        message: impl Into<Cow<'static, str>>,
    ) -> GeneralResult<T> {
        self.map_err(|err| GeneralError {
            status_code,
            message: message.into(),
            error: Some(err.into()),
        })
    }
}

impl<T> ErrorHandlingExt<T, Infallible> for Option<T> {
    fn handle(
        self,
        status_code: StatusCode,
        message: impl Into<Cow<'static, str>>,
    ) -> GeneralResult<T> {
        self.ok_or_else(|| GeneralError {
            status_code,
            message: message.into(),
            error: None,
        })
    }
}
